function [tracks groups] = annotation(img_dir, img_list, H, ...
				      cpoints, groups, ...
				      log_file, old_log_file)
% ANNOTATION   Pedestrian trajectories and groups annotation tool.
%   [new_tracks new_groups] =
%     ANNOTATION(img_dir, img_list, H, cpoints, groups, log_file, old_log_file)
%
%   Input:
%    img_dir [string] Directory where the sequence images are stored.
%    img_list [cell array] List of image file names (without the directory).
%    H [3x3] Sequence ground plane homography (image -> world).
%    cpoints [Nx4] Control point list. Each row contains: track id, x, y,
%       frame, where x and y are in real world coordinates.
%    groups [1xM] Group labels for each id in cpoints (assumes ids = 1:M).
%    log_file [string] Log file name. All actions will be recorded there.
%    old_log_file [string] (optional) When given, the actions recorded in
%       the log file are applied to the cpoints and groups first.
%
%   Output:
%    new_tracks [Kx4] Trajectories interpolated for each frame from annotation.
%    new_groups [1xL] Annotated groups.
%
%   Control (keys/mouse):
%    q - quit
%    a - show/hide annotation
%    A - show/hide complete trajectories of all present people
%
%    left/right arrow - jump 5 frames backward/forward
%    home - jump to the beginning of the sequence
%    end - jump to the end of the sequence
%    shift + left/right arrow - jump 25 frames (1s) backward/forward
%    ctrl + left/right arrow - jump to previous/next control point of the
%       current trajectory
%    ctrl + home/end - jump to first/last control point of the current
%       trajectory
%
%    right click - select the current pedestrian
%    left click - add control point to the current pedestrian's track
%    shift + any click = start new trajectory
%    delete - delete the control point at current frame (if there is any)
%
%    c + right click - 'c' activates connecting mode, right click connects
%       current pedestrian's trajectory with the trajectory of selected
%       pedestrian (pressing 'c' again swithes the mode off again)
%    g + right click - 'g' activates grouping mode, right click groups
%       current pedestrian with the selected one (pressing 'g' again
%       swithes the mode off again)
%    
%   Log file format:
%    The log is a text file which can be freely edited and contains four
%    types of rows:
%     A id, I, J, frame
%      adds a new control point to the annotation of person 'id' at
%      position [I, J] (image coordinates) and frame 'frame'.
%     D id, frame
%      deletes a control point (if present) of person 'id' at frame
%      'frame' from the annotation.
%     C id1, id2
%      connects trajectories of pedestrians 'id1' and 'id2' into a single
%      trajectory with id 'id1' (and merges their group labels accordingly).
%     G id1, id2
%      assignes pedestrians 'id1' and 'id2' into the same group.
%
% (C) 2010, Written by Jan Sochman (Jan.Sochman@cmp.felk.cvut.cz)
% <a href="http://www.comp.leeds.ac.uk/scsjso>School of Computing</a>
% <a href="http://www.leeds.ac.uk>University of Leeds</a>
%

  history_log = [];
  
  fid_log = fopen(log_file, 'w');

  % interpolate cpoints into tracks
  tracks = cpoints2tracks(cpoints);
  
  % process the previous log and perform the prescribed annotation
  if exist('old_log_file', 'var')
    fid_in = fopen(old_log_file, 'r');

    tline = fgetl(fid_in);
    while ischar(tline)
      sp = regexp(tline, '\s|,\s', 'split');
      if sp{1} == 'A'
	add_cpoint([str2num(sp{3}), str2num(sp{4})], str2num(sp{2}), ...
		   str2num(sp{5}));
      elseif sp{1} == 'D'
	delete_cpoint(str2num(sp{2}), str2num(sp{3}));
      elseif sp{1} == 'C'
	connect_tracks(str2num(sp{2}), str2num(sp{3}));
      elseif sp{1} == 'G'
	group_tracks(str2num(sp{2}), str2num(sp{3}));
      end
      tline = fgetl(fid_in);
    end
    
    fclose(fid_in);
  end

  t = 1;	% current frame
  if isempty(tracks)
    frame_idxs = [];
  else
    frame_idxs = find(tracks(:, 4) == t);
  end
  if exist('old_log_file', 'var')
    cur_track = str2num(sp{2});		% lastly edited track
  elseif isempty(tracks)
    cur_track = 0;
  else
    cur_track = min(tracks(:, 1));	% the first track
  end    
  ax = 1;
  last_frame = length(img_list);

  % process control variables
  do_exit = 0;
  do_show_anno = 1;	% show/hide annotation
  do_show_all = 0;	% show complete trajectories of all pedestrians
                        % in the scene
  do_connect = 0;	% used to indicate 'connect' action in progress
  do_group = 0;		% used to indicate 'group' action in progress
  
  init_figure();

  % main loop (waiting for something to happen :)
  prev_t = 1;
  while ~do_exit
    if t ~= prev_t
      if isempty(tracks)
	frame_idxs = [];
      else
	frame_idxs = find(tracks(:, 4) == t);
      end
      show_cur_frame(t);
    end
    prev_t = t;

    pause(0.01);	% pause for callbacks to be called
  end

  % organise the tracks nicely
  tracks = sortrows(tracks, [1 4]);
  
  fclose(fid_log);
  close(fg);

% ~~~~~~~~~~~~~~~~~~~~ END OF THE MAIN PROGRAM ~~~~~~~~~~~~~~~~~~~~~~~~~

% ---- figure_mouseclick ----
function figure_mouseclick(src,eventdata)

  cp = get(gca,'CurrentPoint');

  if cur_track > 0	% modify only when there is something to modify already
    % right button: select an existing track (to switch/group/connect)
    if strcmp(get(src, 'selectiontype'), 'alt') % right button
      pos = world2img(tracks(frame_idxs, 2:3));
      dist = sqrt((pos(:, 1) - cp(1, 2)).^2 + (pos(:, 2) - cp(1, 1)).^2);
      [val, idx] = min(dist);
      if val < 20
	if do_connect			% connect current and selected tracks
	  connect_tracks(cur_track, tracks(frame_idxs(idx), 1));
	  do_connect = 0;
	elseif do_group			% group current and selected tracks
	  group_tracks(cur_track, tracks(frame_idxs(idx), 1));
	  do_group = 0;
	else				% switch to another track
	  cur_track = tracks(frame_idxs(idx), 1);
	end
      end
    end

    % left button: add control point
    if strcmp(get(src, 'selectiontype'), 'normal') % left button
      add_cpoint(cp(1, [2 1]), cur_track, t);
    end
  end

  % shift + any button: create a new track
  if strcmp(get(src, 'selectiontype'), 'extend') % shift + any button
    if cur_track > 0
      cur_track = max(tracks(:, 1)) + 1;
    else
      cur_track = 1;
    end
    add_cpoint(cp(1, [2 1]), cur_track, t);
  end

  % update the view
  show_cur_frame(t);
end

% ---- figure_keypress ----
function figure_keypress(src,evnt)

  if evnt.Character == 'q'
    do_exit = 1;
  % ---------------
  elseif evnt.Character == 'l'
    % show current log (useful e.g. when one is not sure what happend last)
    for I=1:length(history_log)
      disp(history_log{I})
    end
  % ---------------
  elseif evnt.Character == 'a'
    % show/hide annotation
    do_show_anno = 1 - do_show_anno;
    show_cur_frame(t);
  % ---------------
  elseif evnt.Character == 'A'
    % show/hide complete trajectories of the pedestrians in the scene
    do_show_all = 1 - do_show_all;
    show_cur_frame(t);
  % ---------------
  elseif evnt.Character == 'c'
    % start/end 'connect' action (is finished by right click on another person)
    if cur_track > 0
      do_connect = 1 - do_connect;
      show_cur_frame(t);
    end
  % ---------------
  elseif evnt.Character == 'g'
    % start/end 'group' action (is finished by right click on another person)
    if cur_track > 0
      do_group = 1 - do_group;
      show_cur_frame(t);
    end
  % ---------------
  elseif strcmp(evnt.Key, 'delete')
    % delete a control point of selected person at current frame (if any)
    if cur_track > 0
      delete_cpoint(cur_track, t);
      if isempty(cpoints)
	cur_track = 0;
      elseif cur_track > max(cpoints(:, 1))
	cur_track = max(cpoints(:, 1));
      end
      show_cur_frame(t);
    end
  % ---------------
  elseif strcmp(evnt.Key, 'leftarrow')
    % leftarrow: jump 5 frames back
    % shift + leftarrow: jump 25 frames (1s) back
    % ctrl + leftarrow: jump to the previous control point
    step = 5;
    if length(evnt.Modifier) == 1 & strcmp(evnt.Modifier{:},'shift')
      step = 25;
    end
    if length(evnt.Modifier) == 1 & strcmp(evnt.Modifier{:},'control')
      cp = find_prev_cpoint();
      if isempty(cp)
	step = 0;
      else
	step = t - cp(4);
      end
    end
    if t > step
      t = t - step;
    else
      t = 1;
    end
  % ---------------
  elseif strcmp(evnt.Key, 'rightarrow')
    % rightarrow: jump 5 frames forward
    % shift + rightarrow: jump 25 frames (1s) forward
    % ctrl + rightarrow: jump to the next control point
    step = 5;
    if length(evnt.Modifier) == 1 & strcmp(evnt.Modifier{:},'shift')
      step = 25;
    end
    if length(evnt.Modifier) == 1 & strcmp(evnt.Modifier{:},'control')
      cp = find_next_cpoint();
      if isempty(cp)
	step = 0;
      else
	step = cp(4) - t;
      end
    end
    if t + step <= last_frame
      t = t + step;
    else
      t = last_frame;
    end
  % ---------------
  elseif strcmp(evnt.Key, 'end') & length(evnt.Modifier) == 0
    % jump to the end of the sequence
    t = last_frame;
  % ---------------
  elseif strcmp(evnt.Key, 'end') & length(evnt.Modifier) == 1 & ...
	strcmp(evnt.Modifier{:},'control')
    cp = get_last_cpoint();
    if isempty(cp)
      step = 0;
    else
      step = cp(4) - t;
    end
    t = t + step;
  % ---------------
  elseif strcmp(evnt.Key, 'home') & length(evnt.Modifier) == 0
    % jump to the beginning of the sequence
    t = 1;
  elseif strcmp(evnt.Key, 'home') & length(evnt.Modifier) == 1 & ...
	strcmp(evnt.Modifier{:},'control')
    cp = get_first_cpoint();
    if isempty(cp)
      step = 0;
    else
      step = cp(4) - t;
    end
    t = t + step;
  end
end

% ---- init_figure ----
function init_figure()
  % create the fiture
  fg = figure('WindowButtonDownFcn',@figure_mouseclick, ...
	      'WindowKeyPressFcn', @figure_keypress);
  set(fg,'Tag','Annotation:Main');
  set(fg,'units','pixels','position',[150, 150, 700, 550])

  % help panel
  pos = get(fg,'position');
  help_panel = uipanel('units', 'pixels', 'position', ...
		       [10,10, pos(3)-20, 40]);

  % axes
  ax = axes('units','pixels','position',[10,60,pos(3)-20,pos(4)-70],...
	    'visible','off');
  set(fg,'CurrentAxes',ax)
  cla(ax)

  % help text
  pos = get(help_panel,'position');
% $$$   uicontrol('parent', help_panel, 'style', 'text', ...
% $$$ 	    'string', 'right arrow = +1s | right arrow = -1s | q = quit, right click = select active, right click = add ctrl point', ...
% $$$ 	    'position', [5 5 pos(3)-5, pos(4)-10], ...
% $$$ 	    'horizontalalignment', 'left');

  % display current frame
  show_cur_frame(t);
end

% ---- show_cur_frame ----
function show_cur_frame(cur_t)
  % display image + annotation
  img = imread(fullfile(img_dir, img_list{cur_t}));
  cla
  image('parent', ax, 'cdata', img)
  axis(ax, 'image', 'ij')

  if do_show_anno
    show_anno(cur_t);
  end
end

% ---- show_anno ----
function show_anno(cur_t)
  % displays the annotation

  if isempty(tracks)
    return;
  end

  hold on

  frame_idxs = find(tracks(:, 4) == t);
  
  % show complete tracks of present pedestrians (white lines)
  if do_show_all
    for K=1:length(frame_idxs)
      id = tracks(frame_idxs(K), 1);
      cur_pos = world2img(tracks(tracks(:, 1) == id, [2 3]));
      plot(cur_pos(:, 2), cur_pos(:, 1), 'w-');
    end
  end
  
  % draw active pedestrian trajectory (yellow line) and its control
  % points (color codes action in progress)
  cur_pos = world2img(tracks(tracks(:, 1) == cur_track, [2 3]));
  plot(cur_pos(:, 2), cur_pos(:, 1), 'y-');
  cur_pos = world2img(cpoints(cpoints(:, 1) == cur_track, [2 3]));
  if do_connect		% connect action in progress
    plot(cur_pos(:, 2), cur_pos(:, 1), 'm.');
  elseif do_group	% group action in progress
    plot(cur_pos(:, 2), cur_pos(:, 1), 'r.');
  else			% standard annotation in progress
    plot(cur_pos(:, 2), cur_pos(:, 1), 'g.');
  end

  % show groups
  for K=1:length(frame_idxs)
    X = world2img(tracks(frame_idxs(K), 2:3));
    plot(X(2), X(1), '.', 'MarkerSize', 20);

    gid = groups(tracks(frame_idxs(K), 1));
    if gid ~= 0
      friends = find(groups == gid);
      idxs = find(ismember(tracks(frame_idxs, 1), friends));
      Y = world2img(tracks(frame_idxs(idxs), [2 3]));
      for I=1:length(idxs)
	plot([X(2) Y(I, 2)], [X(1) Y(I, 1)], 'r-', 'LineWidth', 3);
      end
    end
  end

  % show IDs
  for K=1:length(frame_idxs)
    X = world2img(tracks(frame_idxs(K), 2:3));
    text(X(2), X(1), num2str(tracks(frame_idxs(K), 1)), ...
	 'Color', [0.99 0.99 0.99]);
  end  
end

% ---- add_cpoint ----
function add_cpoint(cp, id, cur_t)
  % adds new control point 'cp' at frame 'cur_t' to the control point list
  % of pedestrian 'id'

  if ~isempty(cpoints)
    % remove a another control point for the same id and time
    cpoints(find(cpoints(:, 4) == cur_t & cpoints(:, 1) == id), :) = [];
  end

  % if new track has been created
  if id > length(groups)
    groups(end+1) = 0;
  end
  
  % add new control point
  cpoints(end+1, :) = [id, img2world(cp), cur_t];

  % recompute the track
  cps = sortrows(cpoints(cpoints(:, 1) == id, :), 4);
  if ~isempty(tracks)
    tracks(tracks(:, 1) == id, :) = [];
  end
  tracks = [tracks; interp_cpoints(cps)];

  % output to the log
  history_log{end+1} = ['A ' num2str(id) ', ' num2str(cp(1)) ', ' ...
		    num2str(cp(2)) ', ' num2str(cur_t)];
  fprintf(fid_log, '%s\n', history_log{end});
end

% ---- delete_cpoint ----
function delete_cpoint(id, cur_t)
  % removes a control point of pedestrian 'id' at frame 'cur_t'

  % find the control point at frame 'cur_t'
  idx = find(cpoints(:, 1) == id & cpoints(:, 4) == cur_t);

  if isempty(idx)
    return;	% no control point found -> nothing to do
  end
  
  % delete the control point
  cpoints(idx, :) = [];
  if isempty(find(cpoints(:, 1) == id))  % was the the last control point
    groups(id) = [];
    tracks(tracks(:, 1) == id, :) = [];
    if isempty(cpoints)			% very last control point
      tracks = [];
    else
      % set ids to 1:N
      uid = unique(cpoints(:, 1));
      for I=1:length(uid)
	idxs = find(cpoints(:, 1) == uid(I));
	cpoints(idxs, 1) = I;
	idxs = find(tracks(:, 1) == uid(I));
	tracks(idxs, 1) = I;
      end
    end
  else
    % recompute the track
    cps = sortrows(cpoints(cpoints(:, 1) == id, :), 4);
    tracks(tracks(:, 1) == id, :) = [];
    tracks = [tracks; interp_cpoints(cps)];
  end
  
  % output to the log
  history_log{end+1} = ['D ' num2str(id) ', ' num2str(cur_t)];
  fprintf(fid_log, '%s\n', history_log{end});
end

% ---- group_tracks ----
function group_tracks(id1, id2)
  % assign pedestrians 'id1' and 'id2' to the same group
    
  if groups(id1) == 0 & groups(id2) == 0     % none is yet in a group
    groups(id1) = max(groups) + 1;
    groups(id2) = groups(id1);
  elseif groups(id1) == 0 | groups(id2) == 0 % one is already part of a group
    groups(id1) = max(groups(id1), groups(id2));
    groups(id2) = max(groups(id1), groups(id2));
  elseif groups(id1) == groups(id2)	% already in the same group
    % remove id1 from that group
    gid = groups(id1);
    groups(id1) = 0;
    if sum(groups == gid) == 1
      groups(id2) = 0;
      a = find(groups > gid);
      groups(a) = groups(a) - 1;
    end
  else				% both are already in different groups
    mi = min(groups(id1), groups(id2));
    ma = max(groups(id1), groups(id2));
    groups(find(groups == ma)) = mi;
    a = find(groups > ma);
    groups(a) = groups(a) - 1;
  end

  % output to the log
  history_log{end+1} = ['G ' num2str(id1) ', ' num2str(id2)];
  fprintf(fid_log, '%s\n', history_log{end});
end

% ---- connect_tracks ----
function connect_tracks(id1, id2)
  % connects two tracks with ids 'id1' and 'id2' into a single track with
  % id 'id1'

  % assign pedestrians to the same group
  if groups(id1) == 0 & groups(id2) == 0     % none is yet in a group
    % nothing to do
  elseif groups(id1) == 0 | groups(id2) == 0 % one is already part of a group
    groups(id1) = max(groups(id1), groups(id2));
    groups(id2) = max(groups(id1), groups(id2));
  elseif groups(id1) == groups(id2)	% already in the same group
    % nothing to do
  else				% both are already in different groups
    mi = min(groups(id1), groups(id2));
    ma = max(groups(id1), groups(id2));
    groups(find(groups == ma)) = mi;
    a = find(groups > ma);
    groups(a) = groups(a) - 1;
  end

  % id2 -> id1
  cpoints(find(cpoints(:, 1) == id2), 1) = id1;
  groups(id2) = [];

  % recompute the tracks
  tracks(tracks(:, 1) == id1 | tracks(:, 1) == id2, :) = [];
  cps = sortrows(cpoints(cpoints(:, 1) == id1, :), 4);
  tracks = [tracks; interp_cpoints(cps)];

  % set ids to 1:N
  uid = unique(cpoints(:, 1));
  for I=1:length(uid)
    idxs = find(cpoints(:, 1) == uid(I));
    cpoints(idxs, 1) = I;
    idxs = find(tracks(:, 1) == uid(I));
    tracks(idxs, 1) = I;
  end

  % output to the log
  history_log{end+1} = ['C ' num2str(id1) ', ' num2str(id2)];
  fprintf(fid_log, '%s\n', history_log{end});
end

% ---- find_next_cpoint ----
function cp = find_next_cpoint()
  if isempty(cpoints)
    cp = [];
    return;
  end
  
  cps = sortrows(cpoints(cpoints(:, 1) == cur_track, :), 4);
  idx = find(cps(:, 4) > t, 1, 'first');
  if isempty(idx)
    cp = [];
  else
    cp = cps(idx, :);
  end
end

% ---- find_prev_cpoint ----
function cp = find_prev_cpoint()
  if isempty(cpoints)
    cp = [];
    return;
  end
    
  cps = sortrows(cpoints(cpoints(:, 1) == cur_track, :), 4);
  idx = find(cps(:, 4) < t, 1, 'last');
  if isempty(idx)
    cp = [];
  else
    cp = cps(idx, :);
  end
end

% ---- get_first_cpoint ----
function cp = get_first_cpoint()
  if isempty(cpoints)
    cp = [];
    return;
  end
    
  cps = sortrows(cpoints(cpoints(:, 1) == cur_track, :), 4);
  cp = cps(1, :);
end

% ---- get_last_cpoint ----
function cp = get_last_cpoint()
  if isempty(cpoints)
    cp = [];
    return;
  end
    
  cps = sortrows(cpoints(cpoints(:, 1) == cur_track, :), 4);
  cp = cps(end, :);
end

% ---- world2img ----
function Y = world2img(X)
  Y = inv(H) * [X, ones(size(X, 1), 1)]';
  Y = Y ./ [Y(3, :); Y(3, :); Y(3, :)];
  Y = Y(1:2, :)';
end

% ---- img2world ----
function Y = img2world(X)
  Y = H * [X'; ones(1, size(X, 1))];
  Y = Y .* [1./Y(3, :); 1./Y(3, :); 1./Y(3, :)];
  Y = Y(1:2, :)';
end

% ---- cpoints2tracks ----
function tracks = cpoints2tracks(cpoints)
  % interpolates control points into tracks
  
  tracks = [];

  if isempty(cpoints)
    return;
  end
  
  ids = unique(cpoints(:, 1));
  for I =1:length(ids)
    cp = sortrows(cpoints(cpoints(:, 1) == ids(I), :), 4);
    tracks = [tracks; interp_cpoints(cp)];
  end
end

% ---- interp_cpoints ----
function track = interp_cpoints(cp)
  % interpolate the control points
  
  track = [cp(1, :)];
  for J=2:size(cp, 1)
    if cp(J, 4) - cp(J-1, 4) == 1
      track = [track; cp(J, :)];
    else
      % interpolate
      if J > 3
	x = cp(J-2:J, 4);
	Y = cp(J-2:J, [2 3]);
      else
	x = cp(J-1:J, 4);
	Y = cp(J-1:J, [2 3]);
      end
      if J < size(cp, 1)
	x(end+1) = cp(J+1, 4);
	Y(end+1, :) = cp(J+1, [2 3]);
      end
      frames = cp(J-1, 4)+1:cp(J, 4);
      pp = spline(x',Y');
      yy = ppval(pp, frames);
      track = [track; ...
	       cp(1, 1) * ones(length(frames), 1), yy', frames'];
    end
  end
end

end